"""
==========================
Callosal bundles using AFQ API
==========================
An example using the AFQ API to find callosal bundles using the templates from:
http://hdl.handle.net/1773/34926
"""
import os.path as op
import matplotlib.pyplot as plt
import nibabel as nib

import plotly

from AFQ.api.group import GroupAFQ
import AFQ.api.bundle_dict as abd
from AFQ.definitions.image import RoiImage
import AFQ.data.fetch as afd

##########################################################################
# Get some example data
# ---------------------
#
# Retrieves `Stanford HARDI dataset <https://purl.stanford.edu/ng782rw8378>`_.
#

afd.organize_stanford_data(clear_previous_afq="track")

##########################################################################
# Set tractography parameters (optional)
# ---------------------
# We make this tracking_params which we will pass to the GroupAFQ object
# which specifies that we want 100,000 seeds randomly distributed
# in the ROIs of every bundle.
#
# We only do this to make this example faster and consume less space.

tracking_params = dict(seed_mask=RoiImage(),
                       n_seeds=10000,
                       random_seeds=True,
                       rng_seed=42)

##########################################################################
# Set segmentation parameters (optional)
# ---------------------
# We make this segmentation_params which we will pass to the GroupAFQ object
# which specifies that we want to clip the extracted tract profiles
# to only be between the two ROIs.
#
# We do this because tract profiles become less reliable as the bundles
# approach the gray matter-white matter boundary. On some of the non-callosal
# bundles, ROIs are not in a good position to clip edges. In these cases,
# one can remove the first and last nodes in a tract profile.

segmentation_params = {"clip_edges": True}

##########################################################################
# Initialize a GroupAFQ object:
# -------------------------
#
# We specify bundle_info as the callosal bundles only
# (`abd.callosal_bd`). If we want to segment both the callosum
# and the other bundles, we would pass
# `abd.callosal_bd() + abd.default18_bd()`
# instead. This would tell the GroupAFQ object to use bundles from both
# the standard and callosal templates.

myafq = GroupAFQ(
    bids_path=op.join(afd.afq_home, 'stanford_hardi'),
    preproc_pipeline='vistasoft',
    bundle_info=abd.callosal_bd(),
    tracking_params=tracking_params,
    segmentation_params=segmentation_params,
    viz_backend_spec='plotly_no_gif')

# Calling export all produces all of the outputs of processing, including
# tractography, scalar maps, tract profiles and visualizations:
myafq.export_all()


##########################################################################
# Create Group Density Maps:
# -------------------------
#
# pyAFQ can make density maps of streamline counts per subject/session
# by calling `myafq.export("density_map")`. When using `GroupAFQ`, you can also
# combine these into one file by calling `myafq.export_group_density()`.
group_density = myafq.export_group_density()
group_density = nib.load(group_density).get_fdata()
fig, ax = plt.subplots(1)
ax.matshow(
    group_density[:, :, group_density.shape[-1] // 2, 0],
    cmap='viridis')
ax.axis("off")


##########################################################################
# Visualizing bundles and tract profiles:
# ---------------------------------------
# This would run the script and visualize the bundles using the plotly
# interactive visualization, which should automatically open in a
# new browser window.
bundle_html = myafq.export("all_bundles_figure")
plotly.io.show(bundle_html["01"][0])



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

from plotly.io._sg_scraper import plotly_sg_scraper
from AFQ.utils.docs import PNGScraper

import sys
import os
import AFQ

sys.path.insert(0, os.path.abspath('.'))
sys.path.append(os.path.abspath('sphinxext'))


# -- Project information -----------------------------------------------------

project = 'AFQ'
copyright = '2018--, The pyAFQ Contributors'
author = ''

# The short X.Y version
version = ''
# The full version, including alpha/beta/rc tags
release = AFQ.__version__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx_gallery.gen_gallery',
    'sphinx_design',
    'sphinx.ext.autosummary',
    'autoapi.extension',
    'numpydoc',
    'updatedocs',
    'kwargsdocs',
    'methodsdocs',
    'myst_nb',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = 'pydata_sphinx_theme'

# Added theme configuration. See: https://pydata-sphinx-theme.readthedocs.io/
html_logo = "_static/logo.png"

html_sidebars = {
    "**": ["search-field", "sidebar-nav-bs", "globaltoc.html"]
}
html_theme_options = {
    "use_edit_page_button": True,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/yeatmanlab/pyAFQ",
            "icon": "fab fa-github-square",
        }]

}

html_sidebars = {
    "**": ['search-field.html', 'sidebar-nav-bs.html'],
}

html_context = {
    "github_url": "https://github.com",
    "github_user": "yeatmanlab",
    "github_repo": "pyAFQ",
    "github_version": "master",
    "doc_path": "docs/source",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

# html_css_files = [
#     'css/custom.css'
# ]

html_css_files = ['custom.css']

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'pyAFQdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'pyAFQ.tex', 'pyAFQ Documentation',
     'Ariel Rokem', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'pyafq', 'pyAFQ Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'pyAFQ', 'pyAFQ Documentation',
     author, 'pyAFQ', 'One line description of project.',
     'Miscellaneous'),
]


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'python': ('https://docs.python.org/3/', None),
                       'numpy': ('https://docs.scipy.org/doc/numpy/', None),
                       'dipy': ('https://dipy.org/documentation/latest',
                                'https://dipy.org/documentation/1.4.1./objects.inv/')
                       }

image_scrapers = ('matplotlib', plotly_sg_scraper, PNGScraper())

from _progressbars import reset_progressbars  # noqa

sphinx_gallery_thumbnail_path = '../source/_static/logo.png'

from sphinx_gallery.sorting import FileNameSortKey

sphinx_gallery_conf = {

    # path to your examples scripts
    'examples_dirs': ['../../examples/howto_examples',
                      '../../examples/tutorial_examples'],
    # path where to save gallery generated examples
    'gallery_dirs': ['howto/howto_examples', 'tutorials/tutorial_examples'],
    'image_scrapers': image_scrapers,
    'reset_modules': (reset_progressbars),
    'show_memory': True,
    'within_subsection_order': FileNameSortKey,
}

# Auto API
autoapi_type = 'python'
autoapi_dirs = ['../../AFQ']
autoapi_ignore = [
    '*test*', '*_fixes*', '*version*',
    'pyAFQ', 'License', 'afq_bids_entities.json']
autoapi_root = 'reference/api'



import gc
from dipy.io.streamline import load_tractogram
import numpy as np
from dipy.io.stateful_tractogram import StatefulTractogram, Space
import os.path as op


try:
    from trx.io import load as load_trx
    has_trx = True
except ModuleNotFoundError:
    has_trx = False

from AFQ.utils.path import drop_extension, read_json


class SegmentedSFT():
    def __init__(self, bundles, space, sidecar_info={}):
        reference = None
        self.bundle_names = []
        self.sidecar_info = sidecar_info
        sls = []
        idxs = {}
        this_tracking_idxs = []
        idx_count = 0
        for b_name in bundles:
            if isinstance(bundles[b_name], dict):
                this_sls = bundles[b_name]['sl']
                this_tracking_idxs.extend(bundles[b_name]['idx'])
            else:
                this_sls = bundles[b_name]
            if reference is None:
                reference = this_sls
            this_sls = list(this_sls.streamlines)
            sls.extend(this_sls)
            new_idx_count = idx_count + len(this_sls)
            idxs[b_name] = np.arange(idx_count, new_idx_count, dtype=np.uint32)
            idx_count = new_idx_count
            self.bundle_names.append(b_name)

        self.sft = StatefulTractogram(sls, reference, space)
        self.bundle_idxs = idxs
        if len(this_tracking_idxs) > 1:
            self.this_tracking_idxs = this_tracking_idxs
        else:
            self.this_tracking_idxs = None

    def get_sft_and_sidecar(self):
        sidecar_info = {}
        sidecar_info["bundle_ids"] = {}
        dps = np.zeros(len(self.sft.streamlines))
        for ii, bundle_name in enumerate(self.bundle_names):
            sidecar_info["bundle_ids"][f"{bundle_name}"] = ii + 1
            dps[self.bundle_idxs[bundle_name]] = ii + 1
        dps = {"bundle": dps}
        self.sft.data_per_streamline = dps
        if self.this_tracking_idxs is not None:
            for ii in range(len(self.this_tracking_idxs)):
                self.this_tracking_idxs[ii] = int(self.this_tracking_idxs[ii])
            sidecar_info["tracking_idx"] = self.this_tracking_idxs
        return self.sft, sidecar_info

    def get_bundle(self, b_name):
        return self.sft[self.bundle_idxs[b_name]]

    def get_bundle_param_info(self, b_name):
        return self.sidecar_info.get(
            "Bundle Parameters", {}).get(b_name, {})

    @classmethod
    def fromfile(cls, trk_or_trx_file, reference="same", sidecar_file=None):
        if sidecar_file is None:
            # assume json sidecar has the same name as trk_file,
            # but with json suffix
            sidecar_file = f'{drop_extension(trk_or_trx_file)}.json'
            if not op.exists(sidecar_file):
                raise ValueError((
                    "JSON sidecars are required for trk files. "
                    f"JSON sidecar not found for: {sidecar_file}"))
        sidecar_info = read_json(sidecar_file)
        if trk_or_trx_file.endswith(".trx"):
            trx = load_trx(trk_or_trx_file, reference)
            trx.streamlines._data = trx.streamlines._data.astype(np.float32)
            sft = trx.to_sft()
            if reference == "same":
                reference = sft
            bundles = {}
            for bundle in trx.groups:
                idx = trx.groups[bundle]
                bundles[bundle] = StatefulTractogram(
                    sft.streamlines[idx], reference, Space.RASMM)
        else:
            sft = load_tractogram(trk_or_trx_file, reference, Space.RASMM)

            if reference == "same":
                reference = sft
            bundles = {}
            if "bundle_ids" in sidecar_info:
                for b_name, b_id in sidecar_info["bundle_ids"].items():
                    idx = np.where(
                        sft.data_per_streamline['bundle'] == b_id)[0]
                    bundles[b_name] = StatefulTractogram(
                        sft.streamlines[idx], reference, Space.RASMM)
            else:
                bundles["whole_brain"] = sft

        return cls(bundles, Space.RASMM, sidecar_info)


def split_streamline(streamlines, sl_to_split, split_idx):
    """
    Given a Streamlines object, split one of the underlying streamlines

    Parameters
    ----------
    streamlines : a Streamlines class instance
        The group of streamlines, one of which is being split.
    sl_to_split : int
        The index of the streamline that is being split
    split_idx : int
        Where is the streamline being split
    """
    this_sl = streamlines[sl_to_split]

    streamlines._lengths = np.concatenate([
        streamlines._lengths[:sl_to_split],
        np.array([split_idx]),
        np.array([this_sl.shape[0] - split_idx]),
        streamlines._lengths[sl_to_split + 1:]])

    streamlines._offsets = np.concatenate([
        np.array([0]),
        np.cumsum(streamlines._lengths[:-1])])

    return streamlines



import numpy as np
import logging

import nibabel as nib

from dipy.segment.mask import median_otsu
from dipy.align import resample

from AFQ.definitions.utils import Definition, find_file, name_from_path

from skimage.morphology import convex_hull_image, binary_opening

__all__ = [
    "ImageFile", "FullImage", "RoiImage", "B0Image", "LabelledImageFile",
    "ThresholdedImageFile", "ScalarImage", "ThresholdedScalarImage",
    "TemplateImage", "GQImage"]


logger = logging.getLogger('AFQ')


def _resample_image(image_data, dwi_data, image_affine, dwi_affine):
    '''
    Helper function
    Resamples image to dwi if necessary
    '''
    image_type = image_data.dtype
    if ((dwi_data is not None)
        and (dwi_affine is not None)
            and (dwi_data[..., 0].shape != image_data.shape)):
        return np.round(resample(
            image_data.astype(float),
            dwi_data[..., 0],
            image_affine,
            dwi_affine).get_fdata()).astype(image_type)
    else:
        return image_data


class ImageDefinition(Definition):
    '''
    All Image Definitions should inherit this.
    '''

    def get_name(self):
        raise NotImplementedError("Please implement a get_name method")

    def get_image_getter(self, task_name):
        raise NotImplementedError(
            "Please implement a get_image_getter method")


class CombineImageMixin(object):
    """
    Helper Class
    Useful for making an image by combining different conditions
    """

    def __init__(self, combine):
        self.combine = combine.lower()

    def reset_image_draft(self, shape):
        if self.combine == "or":
            self.image_draft = np.zeros(shape, dtype=bool)
        elif self.combine == "and":
            self.image_draft = np.ones(shape, dtype=bool)
        else:
            self.combine_illdefined()

    def __mul__(self, other_image):
        if self.combine == "or":
            return np.logical_or(self.image_draft, other_image)
        elif self.combine == "and":
            return np.logical_and(self.image_draft, other_image)
        else:
            self.combine_illdefined()

    def combine_illdefined(self):
        raise TypeError((
            f"combine should be either 'or' or 'and',"
            f" you set combine to {self.combine}"))


class ImageFile(ImageDefinition):
    """
    Define an image based on a file.
    Does not apply any labels or thresholds;
    Generates image with floating point data.
    Useful for seed and stop images, where threshold can be applied
    after interpolation (see example).

    Parameters
    ----------
    path : str, optional
        path to file to get image from. Use this or suffix.
        Default: None
    suffix : str, optional
        suffix to pass to bids_layout.get() to identify the file.
        Default: None
    filters : str, optional
        Additional filters to pass to bids_layout.get() to identify
        the file.
        Default: {}

    Examples
    --------
    seed_image = ImageFile(
        suffix="WM",
        filters={"scope":"dmriprep"})
    api.GroupAFQ(tracking_params={"seed_image": seed_image,
                                "seed_threshold": 0.1})
    """

    def __init__(self, path=None, suffix=None, filters={}):
        if path is None and suffix is None:
            raise ValueError((
                "One of `path` or `suffix` must set to "
                "a value other than None."))

        if path is not None:
            self._from_path = True
            self.fname = path
        else:
            self._from_path = False
            self.suffix = suffix
            self.filters = filters
            self.fnames = {}

    def find_path(self, bids_layout, from_path,
                  subject, session, required=True):
        if self._from_path:
            return

        nearest_image = find_file(
            bids_layout, from_path, self.filters, self.suffix, session,
            subject, required=required)

        if nearest_image is None:
            return False

        self.fnames[from_path] = nearest_image

    def get_path_data_affine(self, dwi_path):
        if self._from_path:
            image_file = self.fname
        else:
            image_file = self.fnames[dwi_path]
        image_img = nib.load(image_file)
        return image_file, image_img.get_fdata(), image_img.affine

    # This function is set up to be overriden by other images
    def apply_conditions(self, image_data_orig, image_file):
        return image_data_orig, dict(source=image_file)

    def get_name(self):
        return name_from_path(self.fname) if self._from_path else self.suffix

    def get_image_getter(self, task_name):
        def _image_getter_helper(dwi, dwi_data_file):
            # Load data
            image_file, image_data_orig, image_affine = \
                self.get_path_data_affine(dwi_data_file)

            # Apply any conditions on the data
            image_data, meta = self.apply_conditions(
                image_data_orig, image_file)

            # Resample to DWI data:
            image_data = _resample_image(
                image_data,
                dwi.get_fdata(),
                image_affine,
                dwi.affine)
            return nib.Nifti1Image(
                image_data.astype(np.float32),
                dwi.affine), meta
        if task_name == "data":
            def image_getter(dwi, dwi_data_file):
                return _image_getter_helper(dwi, dwi_data_file)
        else:
            def image_getter(data_imap, dwi_data_file):
                return _image_getter_helper(data_imap["dwi"], dwi_data_file)
        return image_getter


class FullImage(ImageDefinition):
    """
    Define an image which covers a full volume.

    Examples
    --------
    brain_image_definition = FullImage()
    """

    def __init__(self):
        pass

    def get_name(self):
        return "entire_volume"

    def get_image_getter(self, task_name):
        def _image_getter_helper(dwi):
            return nib.Nifti1Image(
                np.ones(dwi.get_fdata()[..., 0].shape, dtype=np.float32),
                dwi.affine), dict(source="Entire Volume")
        if task_name == "data":
            def image_getter(dwi):
                return _image_getter_helper(dwi)
        else:
            def image_getter(data_imap):
                return _image_getter_helper(data_imap["dwi"])
        return image_getter


class RoiImage(ImageDefinition):
    """
    Define an image which is all include ROIs or'd together.

    Parameters
    ----------
    use_waypoints : bool
        Whether to use the include ROIs to generate the image.
    use_presegment : bool
        Whether to use presegment bundle dict from segmentation params
        to get ROIs.
    use_endpoints : bool
        Whether to use the endpoints ("start" and "end") to generate
        the image.
    tissue_property : str or None
        Tissue property from `scalars` to multiply the ROI image with.
        Can be useful to limit seed mask to the core white matter.
        Note: this must be a built-in tissue property.
        Default: None
    tissue_property_n_voxel : int or None
        Threshold `tissue_property` to a boolean mask with 
        tissue_property_n_voxel number of voxels set to True.
        Default: None
    tissue_property_threshold : int or None
        Threshold to threshold `tissue_property` if a boolean mask is
        desired. This threshold is interpreted as a percentile.
        Overrides tissue_property_n_voxel.
        Default: None
    Examples
    --------
    seed_image = RoiImage()
    api.GroupAFQ(tracking_params={"seed_image": seed_image})
    """

    def __init__(self,
                 use_waypoints=True,
                 use_presegment=False,
                 use_endpoints=False,
                 tissue_property=None,
                 tissue_property_n_voxel=None,
                 tissue_property_threshold=None):
        self.use_waypoints = use_waypoints
        self.use_presegment = use_presegment
        self.use_endpoints = use_endpoints
        self.tissue_property = tissue_property
        self.tissue_property_n_voxel = tissue_property_n_voxel
        self.tissue_property_threshold = tissue_property_threshold
        if not np.logical_or(self.use_waypoints, np.logical_or(
                self.use_endpoints, self.use_presegment)):
            raise ValueError((
                "One of use_waypoints, use_presegment, "
                "use_endpoints, must be True"))

    def get_name(self):
        return "roi"

    def get_image_getter(self, task_name):
        def _image_getter_helper(mapping,
                                 data_imap, segmentation_params):
            image_data = None
            bundle_dict = data_imap["bundle_dict"]
            if self.use_presegment:
                bundle_dict = \
                    segmentation_params["presegment_bundle_dict"]
            else:
                bundle_dict = bundle_dict

            for bundle_name in bundle_dict:
                bundle_entry = bundle_dict.transform_rois(
                    bundle_name,
                    mapping,
                    data_imap["dwi_affine"])
                rois = []
                if self.use_endpoints:
                    rois.extend(
                        [bundle_entry[end_type] for end_type in
                            ["start", "end"] if end_type in bundle_entry])
                if self.use_waypoints:
                    rois.extend(bundle_entry.get('include', []))
                for roi in rois:
                    warped_roi = roi.get_fdata()
                    if image_data is None:
                        image_data = np.zeros(warped_roi.shape)
                    image_data = np.logical_or(
                        image_data,
                        warped_roi.astype(bool))
            if self.tissue_property is not None:
                tp = nib.load(data_imap[self.tissue_property]).get_fdata()
                image_data = image_data.astype(np.float32) * tp
                if self.tissue_property_threshold is not None:
                    zero_mask = image_data == 0
                    image_data[zero_mask] = np.nan
                    tp_thresh = np.nanpercentile(
                        image_data,
                        100 - self.tissue_property_threshold)
                    image_data[zero_mask] = 0
                    image_data = image_data > tp_thresh
                elif self.tissue_property_n_voxel is not None:
                    tp_thresh = np.sort(image_data.flatten())[
                        -1 - self.tissue_property_n_voxel]
                    image_data = image_data > tp_thresh
            if image_data is None:
                raise ValueError((
                    "BundleDict does not have enough ROIs to generate "
                    f"an ROI Image: {bundle_dict._dict}"))
            return nib.Nifti1Image(
                image_data.astype(np.float32),
                data_imap["dwi_affine"]), dict(source="ROIs")

        if task_name == "data":
            raise ValueError((
                "RoiImage cannot be used in this context, as they"
                "require later derivatives to be calculated"))
        elif task_name == "mapping":
            def image_getter(
                    mapping,
                    data_imap, segmentation_params):
                return _image_getter_helper(
                    mapping,
                    data_imap, segmentation_params)
        else:
            def image_getter(
                    mapping_imap,
                    data_imap, segmentation_params):
                return _image_getter_helper(
                    mapping_imap["mapping"],
                    data_imap, segmentation_params)
        return image_getter


class GQImage(ImageDefinition):
    """
    Threshold the anisotropic diffusion component of the 
    Generalized Q-Sampling Model to generate a brain mask
    which will include the eyes, optic nerve, and cerebrum
    but will exclude most or all of the skull.

    Examples
    --------
    api.GroupAFQ(brain_mask_definition=GQImage())
    """

    def __init__(self):
        pass

    def get_name(self):
        return "GQ"

    def get_image_getter(self, task_name):
        def image_getter_helper(gq_aso):
            gq_aso_img = nib.load(gq_aso)
            gq_aso_data = gq_aso_img.get_fdata()
            ASO_mask = convex_hull_image(
                binary_opening(
                    gq_aso_data > 0.1))

            return nib.Nifti1Image(
                ASO_mask.astype(np.float32),
                gq_aso_img.affine), dict(
                    source=gq_aso,
                    technique="GQ ASO thresholded maps")

        if task_name == "data":
            return image_getter_helper
        else:
            return lambda data_imap: image_getter_helper(
                data_imap["gq_aso"])


class B0Image(ImageDefinition):
    """
    Define an image using b0 and dipy's median_otsu.

    Parameters
    ----------
    median_otsu_kwargs: dict, optional
        Optional arguments to pass into dipy's median_otsu.
        Default: {}

    Examples
    --------
    brain_image_definition = B0Image()
    api.GroupAFQ(brain_image_definition=brain_image_definition)
    """

    def __init__(self, median_otsu_kwargs={}):
        self.median_otsu_kwargs = median_otsu_kwargs

    def get_name(self):
        return "b0"

    def get_image_getter(self, task_name):
        def image_getter_helper(b0):
            mean_b0_img = nib.load(b0)
            mean_b0 = mean_b0_img.get_fdata()
            logger.warning((
                "It is reccomended that you provide a brain mask. "
                "It is provided with the brain_mask_definition argument. "
                "Otherwise, the default brain mask is calculated "
                "by using OTSU on the median-filtered B0 image. "
                "This can be unreliable. "))
            _, image_data = median_otsu(mean_b0, **self.median_otsu_kwargs)
            return nib.Nifti1Image(
                image_data.astype(np.float32),
                mean_b0_img.affine), dict(
                    source=b0,
                    technique="median_otsu applied to b0",
                    median_otsu_kwargs=self.median_otsu_kwargs)
        if task_name == "data":
            return image_getter_helper
        else:
            return lambda data_imap: image_getter_helper(data_imap["b0"])


class LabelledImageFile(ImageFile, CombineImageMixin):
    """
    Define an image based on labels in a file.

    Parameters
    ----------
    path : str, optional
        path to file to get image from. Use this or suffix.
        Default: None
    suffix : str, optional
        suffix to pass to bids_layout.get() to identify the file.
        Default: None
    filters : str, optional
        Additional filters to pass to bids_layout.get() to identify
        the file.
        Default: {}
    inclusive_labels : list of ints, optional
        The labels from the file to include from the boolean image.
        If None, no inclusive labels are applied.
    exclusive_labels : list of ints, optional
        The labels from the file to exclude from the boolean image.
        If None, no exclusive labels are applied.
        Default: None.
    combine : str, optional
        How to combine the boolean images generated by inclusive_labels
        and exclusive_labels. If "and", they will be and'd together.
        If "or", they will be or'd.
        Note: in this class, you will most likely want to either set
        inclusive_labels or exclusive_labels, not both,
        so combine will not matter.
        Default: "or"

    Examples
    --------
    brain_image_definition = LabelledImageFile(
        suffix="aseg",
        filters={"scope": "dmriprep"},
        exclusive_labels=[0])
    api.GroupAFQ(brain_image_definition=brain_image_definition)
    """

    def __init__(self, path=None, suffix=None, filters={},
                 inclusive_labels=None,
                 exclusive_labels=None, combine="or"):
        ImageFile.__init__(self, path, suffix, filters)
        CombineImageMixin.__init__(self, combine)
        self.inclusive_labels = inclusive_labels
        self.exclusive_labels = exclusive_labels

    # overrides ImageFile
    def apply_conditions(self, image_data_orig, image_file):
        # For different sets of labels, extract all the voxels that
        # have any / all of these values:
        self.reset_image_draft(image_data_orig.shape)
        if self.inclusive_labels is not None:
            for label in self.inclusive_labels:
                self.image_draft = self * (image_data_orig == label)
        if self.exclusive_labels is not None:
            for label in self.exclusive_labels:
                self.image_draft = self * (image_data_orig != label)

        meta = dict(source=image_file,
                    inclusive_labels=self.inclusive_labels,
                    exclusive_lavels=self.exclusive_labels,
                    combined_with=self.combine)
        return self.image_draft, meta


class ThresholdedImageFile(ImageFile, CombineImageMixin):
    """
    Define an image based on thresholding a file.
    Note that this should not be used to directly make a seed image
    or a stop image. In those cases, consider thresholding after
    interpolation, as in the example for ImageFile.

    Parameters
    ----------
    path : str, optional
        path to file to get image from. Use this or suffix.
        Default: None
    suffix : str, optional
        suffix to pass to bids_layout.get() to identify the file.
        Default: None
    filters : str, optional
        Additional filters to pass to bids_layout.get() to identify
        the file.
        Default: {}
    lower_bound : float, optional
        Lower bound to generate boolean image from data in the file.
        If None, no lower bound is applied.
        Default: None.
    upper_bound : float, optional
        Upper bound to generate boolean image from data in the file.
        If None, no upper bound is applied.
        Default: None.
    as_percentage : bool, optional
        Interpret lower_bound and upper_bound as percentages of the
        total non-nan voxels in the image to include (between 0 and 100),
        instead of as a threshold on the values themselves.
        Default: False
    combine : str, optional
        How to combine the boolean images generated by lower_bound
        and upper_bound. If "and", they will be and'd together.
        If "or", they will be or'd.
        Default: "and"

    Examples
    --------
    brain_image_definition = ThresholdedImageFile(
        suffix="BM",
        filters={"scope":"dmriprep"},
        lower_bound=0.1)
    api.GroupAFQ(brain_image_definition=brain_image_definition)
    """

    def __init__(self, path=None, suffix=None, filters={}, lower_bound=None,
                 upper_bound=None, as_percentage=False, combine="and"):
        ImageFile.__init__(self, path, suffix, filters)
        CombineImageMixin.__init__(self, combine)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.as_percentage = as_percentage

    # overrides ImageFile
    def apply_conditions(self, image_data_orig, image_file):
        # Apply thresholds
        self.reset_image_draft(image_data_orig.shape)
        if self.upper_bound is not None:
            if self.as_percentage:
                upper_bound = np.nanpercentile(
                    image_data_orig,
                    self.upper_bound)
            else:
                upper_bound = self.upper_bound
            self.image_draft = self * (image_data_orig < upper_bound)
        if self.lower_bound is not None:
            if self.as_percentage:
                lower_bound = np.nanpercentile(
                    image_data_orig,
                    100 - self.lower_bound)
            else:
                lower_bound = self.lower_bound
            self.image_draft = self * (image_data_orig > lower_bound)

        meta = dict(source=image_file,
                    upper_bound=self.upper_bound,
                    lower_bound=self.lower_bound,
                    combined_with=self.combine)
        return self.image_draft, meta


class ScalarImage(ImageDefinition):
    """
    Define an image based on a scalar.
    Does not apply any labels or thresholds;
    Generates image with floating point data.
    Useful for seed and stop images, where threshold can be applied
    after interpolation (see example).

    Parameters
    ----------
    scalar : str
        Scalar to threshold.
        Can be one of "dti_fa", "dti_md", "dki_fa", "dki_md".

    Examples
    --------
    seed_image = ScalarImage(
        "dti_fa")
    api.GroupAFQ(tracking_params={
        "seed_image": seed_image,
        "seed_threshold": 0.2})
    """

    def __init__(self, scalar):
        self.scalar = scalar

    def get_name(self):
        return self.scalar

    def get_image_getter(self, task_name):
        if task_name == "data":
            raise ValueError((
                "ScalarImage cannot be used in this context, as they"
                "require later derivatives to be calculated"))

        def image_getter(data_imap):
            return nib.load(data_imap[self.scalar]), dict(
                FromScalar=self.scalar)
        return image_getter


class ThresholdedScalarImage(ThresholdedImageFile, ScalarImage):
    """
    Define an image based on thresholding a scalar image.
    Note that this should not be used to directly make a seed image
    or a stop image. In those cases, consider thresholding after
    interpolation, as in the example for ScalarImage.

    Parameters
    ----------
    scalar : str
        Scalar to threshold.
        Can be one of "dti_fa", "dti_md", "dki_fa", "dki_md".
    lower_bound : float, optional
        Lower bound to generate boolean image from data in the file.
        If None, no lower bound is applied.
        Default: None.
    upper_bound : float, optional
        Upper bound to generate boolean image from data in the file.
        If None, no upper bound is applied.
        Default: None.
    combine : str, optional
        How to combine the boolean images generated by lower_bound
        and upper_bound. If "and", they will be and'd together.
        If "or", they will be or'd.
        Default: "and"

    Examples
    --------
    seed_image = ThresholdedScalarImage(
        "dti_fa",
        lower_bound=0.2)
    api.GroupAFQ(tracking_params={"seed_image": seed_image})
    """

    def __init__(self, scalar, lower_bound=None, upper_bound=None,
                 combine="and"):
        self.scalar = scalar
        CombineImageMixin.__init__(self, combine)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound


class PFTImage(ImageDefinition):
    """
    Define an image for use in PFT tractography. Only use
    if tracker set to 'pft' in tractography.

    Parameters
    ----------
    WM_probseg : ImageFile
        White matter segmentation file.
    GM_probseg : ImageFile
        Gray matter segmentation file.
    CSF_probseg : ImageFile
        Corticospinal fluid segmentation file.

    Examples
    --------
    stop_image = PFTImage(
        afm.ImageFile(suffix="WMprobseg"),
        afm.ImageFile(suffix="GMprobseg"),
        afm.ImageFile(suffix="CSFprobseg"))
    api.GroupAFQ(tracking_params={
        "stop_image": stop_image,
        "stop_threshold": "CMC",
        "tracker": "pft"})
    """

    def __init__(self, WM_probseg, GM_probseg, CSF_probseg):
        self.probsegs = (WM_probseg, GM_probseg, CSF_probseg)

    def find_path(self, bids_layout, from_path,
                  subject, session, required=True):
        if required == False:
            raise ValueError(
                "PFTImage cannot be used in this context")
        for probseg in self.probsegs:
            probseg.find_path(
                bids_layout, from_path, subject, session,
                required=required)

    def get_name(self):
        return "pft"

    def get_image_getter(self, task_name):
        if task_name == "data":
            raise ValueError("PFTImage cannot be used in this context")
        return [probseg.get_image_getter(task_name)
                for probseg in self.probsegs]


class TemplateImage(ImageDefinition):
    """
    Define a scalar based on a template.
    This template will be transformed into subject space before use.

    Parameters
    ----------
    path : str
        path to the template.

    Examples
    --------
    my_scalar = TemplateImage(
        "path/to/my_scalar_in_MNI.nii.gz")
    api.GroupAFQ(scalars=["dti_fa", "dti_md", my_scalar])
    """

    def __init__(self, path):
        self.path = path

    def get_name(self):
        return name_from_path(self.path)

    def get_image_getter(self, task_name):
        def _image_getter_helper(mapping, reg_template, reg_subject):
            img = nib.load(self.path)
            img_data = resample(
                img.get_fdata(),
                reg_template,
                img.affine,
                reg_template.affine).get_fdata()

            scalar_data = mapping.transform_inverse(
                img_data, interpolation='nearest')
            return nib.Nifti1Image(
                scalar_data.astype(np.float32),
                reg_subject.affine), dict(source=self.path)

        if task_name == "data":
            raise ValueError((
                "TemplateImage cannot be used in this context, as they"
                "require later derivatives to be calculated"))
        elif task_name == "mapping":
            def image_getter(mapping, reg_subject, data_imap):
                return _image_getter_helper(
                    mapping, data_imap["reg_template"],
                    reg_subject)
        else:
            def image_getter(mapping_imap, data_imap):
                return _image_getter_helper(
                    mapping_imap["mapping"], data_imap["reg_template"],
                    mapping_imap["reg_subject"])
        return image_getter



import nibabel as nib
import numpy as np
import logging
from time import time
import os.path as op

from AFQ.definitions.utils import Definition, find_file
from dipy.align import syn_registration, affine_registration
import AFQ.registration as reg
from AFQ.utils.path import write_json
from AFQ.tasks.utils import get_fname

from dipy.align.imaffine import AffineMap

try:
    from fsl.data.image import Image
    from fsl.transform.fnirt import readFnirt
    from fsl.transform.nonlinear import applyDeformation
    has_fslpy = True
except ModuleNotFoundError:
    has_fslpy = False

__all__ = ["FnirtMap", "SynMap", "SlrMap", "AffMap", "IdentityMap"]


logger = logging.getLogger('AFQ')


# For map defintions, get_for_subses should return only the mapping
# Where the mapping has transform and transform_inverse functions
# which each accept data, **kwargs


class FnirtMap(Definition):
    """
    Use an existing FNIRT map. Expects a warp file
    and an image file for each subject / session; image file
    is used as src space for warp.

    Parameters
    ----------
    warp_path : str, optional
        path to file to get warp from. Use this or warp_suffix.
        Default: None
    space_path : str, optional
        path to file to get warp from. Use this or space_suffix.
        Default: None
    warp_suffix : str, optional
        suffix to pass to bids_layout.get() to identify the warp file.
        Default: None
    space_suffix : str, optional
        suffix to pass to bids_layout.get() to identify the space file.
        Default: None
    warp_filters : str, optional
        Additional filters to pass to bids_layout.get() to identify
        the warp file.
        Default: {}
    space_filters : str, optional
        Additional filters to pass to bids_layout.get() to identify
        the space file.
        Default: {}

    Notes
    -----
    If you have an existing mapping calculated using Fnirt,
    you can pass bids filters to :class:`AFQ.definitions.mapping.FnirtMap`
    and pyAFQ will find and use that mapping.

    Examples
    --------
    fnirt_map = FnirtMap(
        warp_suffix="warp",
        space_suffix="MNI",
        warp_filters={"scope": "TBSS"},
        space_filters={"scope": "TBSS"})
    api.GroupAFQ(mapping=fnirt_map)
    """

    def __init__(self, warp_path=None, space_path=None,
                 warp_suffix=None, space_suffix=None,
                 warp_filters={}, space_filters={}):
        if not has_fslpy:
            raise ImportError(
                "Please install fslpy if you want to use FnirtMap")
        if warp_path is None and warp_suffix is None:
            raise ValueError((
                "One of `warp_path` or `warp_suffix` should be set "
                "to a value other than None."))
        if space_path is None and space_suffix is None:
            raise ValueError(
                "One of space_path or space_suffix must not be None.")
        if warp_path is not None and space_path is None\
                or space_path is not None and warp_path is None:
            raise ValueError((
                "If passing a value for `warp_path`, "
                "you must also pass a value for `space_path`"))
        if warp_path is not None:
            self._from_path = True
            self.fnames = (warp_path, space_path)
        else:
            self._from_path = False
            self.warp_suffix = warp_suffix
            self.warp_filters = warp_filters
            self.space_suffix = space_suffix
            self.space_filters = space_filters
            self.fnames = {}

    def find_path(self, bids_layout, from_path,
                  subject, session, required=True):
        if self._from_path:
            return
        if session not in self.fnames:
            self.fnames[session] = {}

        nearest_warp = find_file(
            bids_layout, from_path, self.warp_filters, self.warp_suffix,
            session, subject, required=required)

        nearest_space = find_file(
            bids_layout, from_path, self.space_filters, self.space_suffix,
            session, subject, required=required)

        self.fnames[from_path] = (nearest_warp, nearest_space)

    def get_for_subses(self, base_fname, dwi, dwi_data_file, reg_subject,
                       reg_template):
        if self._from_path:
            nearest_warp, nearest_space = self.fnames
        else:
            nearest_warp, nearest_space = self.fnames[dwi_data_file]

        our_templ = reg_template
        subj = Image(dwi)
        their_templ = Image(nearest_space)
        warp = readFnirt(nearest_warp, their_templ, subj)

        return ConformedFnirtMapping(warp, our_templ.affine)


class ConformedFnirtMapping():
    """
        ConformedFnirtMapping which matches the generic mapping API.
    """

    def __init__(self, warp, ref_affine):
        self.ref_affine = ref_affine
        self.warp = warp

    def transform_inverse(self, data, **kwargs):
        data_img = Image(nib.Nifti1Image(
            data.astype(np.float32), self.ref_affine))
        return np.asarray(applyDeformation(data_img, self.warp).data)

    def transform_inverse_pts(self, pts):
        # This should only be used for curvature analysis,
        # Because I think the results still need to be shifted
        pts = nib.affines.apply_affine(
            self.warp.src.getAffine('voxel', 'world'), pts)
        pts = nib.affines.apply_affine(
            np.linalg.inv(self.ref_affine), pts)
        pts = self.warp.transform(pts, 'fsl', "world")
        return pts

    def transform(self, data, **kwargs):
        raise NotImplementedError(
            "Fnirt based mappings can currently"
            + " only transform from template to subject space")


class IdentityMap(Definition):
    """
    Does not perform any transformations from MNI to subject where
    pyAFQ normally would.

    Examples
    --------
    my_example_mapping = IdentityMap()
    api.GroupAFQ(mapping=my_example_mapping)
    """

    def __init__(self):
        pass

    def get_for_subses(self, base_fname, dwi, dwi_data_file, reg_subject,
                       reg_template):
        return ConformedAffineMapping(
            np.identity(4),
            domain_grid_shape=reg.reduce_shape(
                reg_subject.shape),
            domain_grid2world=reg_subject.affine,
            codomain_grid_shape=reg.reduce_shape(
                reg_template.shape),
            codomain_grid2world=reg_template.affine)


class GeneratedMapMixin(object):
    """
    Helper Class
    Useful for maps that are generated by pyAFQ
    """

    def get_fnames(self, extension, base_fname):
        mapping_file = get_fname(
            base_fname,
            '_desc-mapping_from-DWI_to-MNI_xform')
        meta_fname = f'{mapping_file}.json'
        mapping_file = mapping_file + extension
        return mapping_file, meta_fname

    def prealign(self, base_fname, reg_subject, reg_template, save=True):
        prealign_file_desc = "_desc-prealign_from-DWI_to-MNI_xform"
        prealign_file = get_fname(
            base_fname, f'{prealign_file_desc}.npy')
        if not op.exists(prealign_file):
            start_time = time()
            _, aff = affine_registration(
                reg_subject,
                reg_template,
                **self.affine_kwargs)
            meta = dict(
                type="rigid",
                dependent="dwi",
                timing=time() - start_time)
            if not save:
                return aff
            logger.info(f"Saving {prealign_file}")
            np.save(prealign_file, aff)
            meta_fname = get_fname(
                base_fname, f'{prealign_file_desc}.json')
            write_json(meta_fname, meta)
        return prealign_file if save else np.load(prealign_file)

    def get_for_subses(self, base_fname, dwi, dwi_data_file, reg_subject,
                       reg_template, subject_sls=None, template_sls=None):
        mapping_file, meta_fname = self.get_fnames(
            self.extension, base_fname)

        if self.use_prealign:
            reg_prealign = np.load(self.prealign(
                base_fname, reg_subject, reg_template))
        else:
            reg_prealign = None
        if not op.exists(mapping_file):
            start_time = time()
            mapping = self.gen_mapping(
                base_fname, reg_subject, reg_template,
                subject_sls, template_sls,
                reg_prealign)
            total_time = time() - start_time

            logger.info(f"Saving {mapping_file}")
            reg.write_mapping(mapping, mapping_file)
            meta = dict(
                type="displacementfield",
                timing=total_time)
            if subject_sls is None:
                meta["dependent"] = "dwi"
            else:
                meta["dependent"] = "trk"
            if isinstance(reg_subject, str):
                meta["reg_subject"] = reg_subject
            if isinstance(reg_template, str):
                meta["reg_template"] = reg_template
            write_json(meta_fname, meta)
        reg_prealign_inv = np.linalg.inv(reg_prealign) if self.use_prealign\
            else None
        mapping = reg.read_mapping(
            mapping_file,
            dwi,
            reg_template,
            prealign=reg_prealign_inv)
        return mapping


class SynMap(GeneratedMapMixin, Definition):
    """
    Calculate a Syn registration for each subject/session
    using reg_subject and reg_template.

    Parameters
    ----------
    use_prealign : bool
        Whether to perform a linear pre-registration.
        Default: True
    affine_kwargs : dictionary, optional
        Parameters to pass to affine_registration
        in dipy.align, which does the linear pre-alignment.
        Only used if use_prealign is True.
        Default: {}
    syn_kwargs : dictionary, optional
        Parameters to pass to syn_registration
        in dipy.align, which does the SyN alignment.
        Default: {}

    Notes
    -----
    The default mapping class is to
    use Symmetric Diffeomorphic Image Registration (SyN).
    This is done with an optional linear pre-alignment by default.
    The parameters of the pre-alginment can be specified when
    initializing the SynMap.

    Examples
    --------
    api.GroupAFQ(mapping=SynMap())
    """

    def __init__(self, use_prealign=True, affine_kwargs={}, syn_kwargs={}):
        self.use_prealign = use_prealign
        self.affine_kwargs = affine_kwargs
        self.syn_kwargs = syn_kwargs
        self.extension = ".nii.gz"

    def gen_mapping(self, base_fname, reg_subject, reg_template,
                    subject_sls, template_sls,
                    reg_prealign):
        _, mapping = syn_registration(
            reg_subject.get_fdata(),
            reg_template.get_fdata(),
            moving_affine=reg_subject.affine,
            static_affine=reg_template.affine,
            prealign=reg_prealign,
            **self.syn_kwargs)
        if self.use_prealign:
            mapping.codomain_world2grid = np.linalg.inv(reg_prealign)
        return mapping


class SlrMap(GeneratedMapMixin, Definition):
    """
    Calculate a SLR registration for each subject/session
    using reg_subject and reg_template.

    Parameters
    ----------
    slr_kwargs : dictionary, optional
        Parameters to pass to whole_brain_slr
        in dipy, which does the SLR alignment.
        Default: {}

    Notes
    -----
    Use this class to tell pyAFQ to use
    Streamline-based Linear Registration (SLR) 
    for registration. Note that the reg_template and reg_subject
    parameters passed to :class:`AFQ.api.group.GroupAFQ` should
    be streamlines when using this registration.

    Examples
    --------
    api.GroupAFQ(mapping=SlrMap())
    """

    def __init__(self, slr_kwargs={}):
        self.slr_kwargs = {}
        self.use_prealign = False
        self.extension = ".npy"

    def gen_mapping(self, base_fname, reg_template, reg_subject,
                    subject_sls, template_sls, reg_prealign):
        return reg.slr_registration(
            subject_sls, template_sls,
            moving_affine=reg_subject.affine,
            moving_shape=reg_subject.shape,
            static_affine=reg_template.affine,
            static_shape=reg_template.shape,
            **self.slr_kwargs)


class AffMap(GeneratedMapMixin, Definition):
    """
    Calculate an affine registration for each subject/session
    using reg_subject and reg_template.

    Parameters
    ----------
    affine_kwargs : dictionary, optional
        Parameters to pass to affine_registration
        in dipy.align, which does the linear pre-alignment.
        Default: {}

    Notes
    -----
    This will only perform a linear alignment for registration.

    Examples
    --------
    api.GroupAFQ(mapping=AffMap())
    """

    def __init__(self, affine_kwargs={}):
        self.use_prealign = False
        self.affine_kwargs = affine_kwargs
        self.extension = ".npy"

    def gen_mapping(self, base_fname, reg_subject, reg_template,
                    subject_sls, template_sls,
                    reg_prealign):
        return ConformedAffineMapping(
            np.linalg.inv(self.prealign(
                base_fname, reg_subject, reg_template, save=False)),
            domain_grid_shape=reg.reduce_shape(
                reg_subject.shape),
            domain_grid2world=reg_subject.affine,
            codomain_grid_shape=reg.reduce_shape(
                reg_template.shape),
            codomain_grid2world=reg_template.affine)


class ConformedAffineMapping(AffineMap):
    """
    Modifies AffineMap API to match DiffeomorphicMap API.
    Important for SLR maps API to be indistinguishable from SYN maps API.
    """

    def transform(self, *args, interpolation='linear', **kwargs):
        kwargs['interp'] = interpolation
        return super().transform_inverse(*args, **kwargs)

    def transform_inverse(self, *args, interpolation='linear', **kwargs):
        kwargs['interp'] = interpolation
        return super().transform(*args, **kwargs)



import os.path as op
import numpy as np
import numpy.testing as npt
import pytest

from bids.layout import BIDSLayout

import AFQ.definitions.image as afm
from AFQ.definitions.image import *  # interprets images from eval
from AFQ.definitions.mapping import *  # interprets mappings from eval
from AFQ.tests.test_api import create_dummy_bids_path


def test_str_instantiates_mixin():
    thresh_image = afm.ThresholdedScalarImage("dti_fa", lower_bound=0.2)
    thresh_image_str = thresh_image.str_for_toml()
    thresh_image_from_str = eval(thresh_image_str)

    npt.assert_(thresh_image.combine == thresh_image_from_str.combine)
    npt.assert_(thresh_image.lower_bound ==
                thresh_image_from_str.lower_bound)
    npt.assert_(thresh_image.upper_bound ==
                thresh_image_from_str.upper_bound)


def test_resample_image():
    image_data = np.zeros((2, 2, 2), dtype=bool)
    image_data[0] = True
    dwi_data = np.zeros((2, 2, 2, 5))
    image_affine = np.eye(4)
    dwi_affine = np.eye(4) * 2
    npt.assert_array_equal(
        afm._resample_image(image_data, None, image_affine, dwi_affine),
        image_data)
    npt.assert_array_equal(
        afm._resample_image(image_data, dwi_data, image_affine, dwi_affine),
        image_data)

    image_data = np.zeros((3, 3, 3), dtype=bool)
    image_data[0] = True
    resampled_image = afm._resample_image(
        image_data, dwi_data, image_affine, dwi_affine)
    npt.assert_array_equal(
        resampled_image.shape,
        dwi_data[..., 0].shape)
    npt.assert_equal(
        resampled_image.dtype,
        image_data.dtype)


@pytest.mark.parametrize("subject", ["01", "02"])
@pytest.mark.parametrize("session", ["01", "02"])
def test_find_path(subject, session):
    bids_dir = create_dummy_bids_path(2, 2)
    bids_layout = BIDSLayout(bids_dir, derivatives=True)

    test_dwi_path = bids_layout.get(
        subject=subject, session=session, return_type="filename",
        suffix="dwi", extension="nii.gz"
    )[0]

    image_file = ImageFile(suffix="seg", filters={'scope': 'synthetic'})
    image_file.find_path(bids_layout, test_dwi_path, subject, session)

    assert image_file.fnames[test_dwi_path] == op.join(
        bids_dir, "derivatives", "dmriprep", "sub-" + subject,
        "ses-" + session, "anat", "seg.nii.gz"
    )

    other_sub = "01" if subject == "02" else "02"
    with pytest.raises(ValueError):
        image_file.find_path(
            bids_layout,
            test_dwi_path,
            subject=other_sub,
            session=session,
        )



from collections.abc import Iterable
import numpy as np
import nibabel as nib
import logging
from tqdm import tqdm

import dipy.data as dpd
from dipy.align import resample
from dipy.direction import (DeterministicMaximumDirectionGetter,
                            ProbabilisticDirectionGetter)
from dipy.io.stateful_tractogram import StatefulTractogram, Space
from dipy.tracking.stopping_criterion import (ThresholdStoppingCriterion,
                                              CmcStoppingCriterion,
                                              ActStoppingCriterion)

from nibabel.streamlines.tractogram import LazyTractogram

from dipy.tracking.local_tracking import (LocalTracking,
                                          ParticleFilteringTracking)
from AFQ._fixes import tensor_odf
from AFQ.tractography.utils import gen_seeds, get_percentile_threshold


def track(params_file, directions="prob", max_angle=30., sphere=None,
          seed_mask=None, seed_threshold=0, thresholds_as_percentages=False,
          n_seeds=1, random_seeds=False, rng_seed=None, stop_mask=None,
          stop_threshold=0, step_size=0.5, minlen=50, maxlen=250,
          odf_model="CSD", tracker="local", trx=False):
    """
    Tractography

    Parameters
    ----------
    params_file : str, nibabel img.
        Full path to a nifti file containing CSD spherical harmonic
        coefficients, or nibabel img with model params.
    directions : str
        How tracking directions are determined.
        One of: {"det" | "prob"}
        Default: "prob"
    max_angle : float, optional.
        The maximum turning angle in each step. Default: 30
    sphere : Sphere object, optional.
        The discretization of direction getting. default:
        dipy.data.default_sphere.
    seed_mask : array, optional.
        Float or binary mask describing the ROI within which we seed for
        tracking.
        Default to the entire volume (all ones).
    seed_threshold : float, optional.
        A value of the seed_mask above which tracking is seeded.
        Default to 0.
    n_seeds : int or 2D array, optional.
        The seeding density: if this is an int, it is is how many seeds in each
        voxel on each dimension (for example, 2 => [2, 2, 2]). If this is a 2D
        array, these are the coordinates of the seeds. Unless random_seeds is
        set to True, in which case this is the total number of random seeds
        to generate within the mask. Default: 1
    random_seeds : bool
        Whether to generate a total of n_seeds random seeds in the mask.
        Default: False.
    rng_seed : int
        random seed used to generate random seeds if random_seeds is
        set to True. Default: None
    stop_mask : array or str, optional.
        If array: A float or binary mask that determines a stopping criterion
        (e.g. FA).
        If tuple: it contains a sequence that is interpreted as:
        (pve_wm, pve_gm, pve_csf), each item of which is either a string
        (full path) or a nibabel img to be used in particle filtering
        tractography.
        A tuple is required if tracker is set to "pft".
        Defaults to no stopping (all ones).
    stop_threshold : float or tuple, optional.
        If float, this a value of the stop_mask below which tracking is
        terminated (and stop_mask has to be an array).
        If str, "CMC" for Continuous Map Criterion [Girard2014]_.
                "ACT" for Anatomically-constrained tractography [Smith2012]_.
        A string is required if the tracker is set to "pft".
        Defaults to 0 (this means that if no stop_mask is passed,
        we will stop only at the edge of the image).
    thresholds_as_percentages : bool, optional
        Interpret seed_threshold and stop_threshold as percentages of the
        total non-nan voxels in the seed and stop mask to include
        (between 0 and 100), instead of as a threshold on the
        values themselves.
        Default: False
    step_size : float, optional.
        The size of a step (in mm) of tractography. Default: 0.5
    minlen: int, optional
        The miminal length (mm) in a streamline. Default: 20
    maxlen: int, optional
        The miminal length (mm) in a streamline. Default: 250
    odf_model : str, optional
        One of {"DTI", "CSD", "DKI"}. Defaults to use "DTI"
    tracker : str, optional
        Which strategy to use in tracking. This can be the standard local
        tracking ("local") or Particle Filtering Tracking ([Girard2014]_).
        One of {"local", "pft"}. Default: "local"
    trx : bool, optional
        Whether to return the streamlines compatible with input to TRX file
        (i.e., as a LazyTractogram class instance).
        Default: False

    Returns
    -------
    list of streamlines ()

    References
    ----------
    .. [Girard2014] Girard, G., Whittingstall, K., Deriche, R., &
        Descoteaux, M. Towards quantitative connectivity analysis: reducing
        tractography biases. NeuroImage, 98, 266-278, 2014.
    """
    logger = logging.getLogger('AFQ')

    logger.info("Loading Image...")
    if isinstance(params_file, str):
        params_img = nib.load(params_file)
    else:
        params_img = params_file

    model_params = params_img.get_fdata()
    odf_model = odf_model.upper()
    directions = directions.lower()

    # We need to calculate the size of a voxel, so we can transform
    # from mm to voxel units:
    minlen = int(minlen / step_size)
    maxlen = int(maxlen / step_size)

    seeds = gen_seeds(
        seed_mask, seed_threshold,
        n_seeds, thresholds_as_percentages,
        random_seeds, rng_seed, params_img.affine)

    if sphere is None:
        sphere = dpd.default_sphere

    logger.info("Getting Directions...")
    if directions == "det":
        dg = DeterministicMaximumDirectionGetter
    elif directions == "prob":
        dg = ProbabilisticDirectionGetter
    else:
        raise ValueError(f"Unrecognized direction '{directions}'.")

    if odf_model == "DTI" or odf_model == "DKI":
        evals = model_params[..., :3]
        evecs = model_params[..., 3:12].reshape(params_img.shape[:3] + (3, 3))
        odf = tensor_odf(evals, evecs, sphere)
        dg = dg.from_pmf(odf, max_angle=max_angle, sphere=sphere)
    else:
        dg = dg.from_shcoeff(model_params, max_angle=max_angle, sphere=sphere)

    if tracker == "local":
        if stop_mask is None:
            stop_mask = np.ones(params_img.shape[:3])

        if len(np.unique(stop_mask)) <= 2:
            stopping_criterion = ThresholdStoppingCriterion(stop_mask,
                                                            0.5)
        else:
            if thresholds_as_percentages:
                stop_threshold = get_percentile_threshold(
                    stop_mask, stop_threshold)
            stop_mask_copy = np.copy(stop_mask)
            stop_thresh_copy = np.copy(stop_threshold)
            stopping_criterion = ThresholdStoppingCriterion(stop_mask_copy,
                                                            stop_thresh_copy)

        my_tracker = LocalTracking

    elif tracker == "pft":
        if not isinstance(stop_threshold, str):
            raise RuntimeError(
                "You are using PFT tracking, but did not provide a string ",
                "'stop_threshold' input. ",
                "Possible inputs are: 'CMC' or 'ACT'")
        if not (isinstance(stop_mask, Iterable) and len(stop_mask) == 3):
            raise RuntimeError(
                "You are using PFT tracking, but did not provide a length "
                "3 iterable for `stop_mask`. "
                "Expected a (pve_wm, pve_gm, pve_csf) tuple.")
        pves = []
        pve_imgs = []
        vox_sizes = []
        for ii, pve in enumerate(stop_mask):
            if isinstance(pve, str):
                img = nib.load(pve)
            else:
                img = pve
            pve_imgs.append(img)
            pves.append(pve_imgs[-1].get_fdata())
        average_voxel_size = np.mean(vox_sizes)
        pve_wm_img, pve_gm_img, pve_csf_img = pve_imgs
        pve_wm_data, pve_gm_data, pve_csf_data = pves
        pve_wm_data = resample(
            pve_wm_data, model_params[..., 0],
            pve_wm_img.affine,
            params_img.affine).get_fdata()
        pve_gm_data = resample(
            pve_gm_data, model_params[..., 0],
            pve_gm_img.affine,
            params_img.affine).get_fdata()
        pve_csf_data = resample(
            pve_csf_data, model_params[..., 0],
            pve_csf_img.affine,
            params_img.affine).get_fdata()

        vox_sizes.append(np.mean(params_img.header.get_zooms()[:3]))

        my_tracker = ParticleFilteringTracking
        if stop_threshold == "CMC":
            stopping_criterion = CmcStoppingCriterion.from_pve(
                pve_wm_data,
                pve_gm_data,
                pve_csf_data,
                step_size=step_size,
                average_voxel_size=average_voxel_size)
        elif stop_threshold == "ACT":
            stopping_criterion = ActStoppingCriterion.from_pve(
                pve_wm_data,
                pve_gm_data,
                pve_csf_data)

    logger.info(
        f"Tracking with {len(seeds)} seeds, 2 directions per seed...")

    return _tracking(my_tracker, seeds, dg, stopping_criterion, params_img,
                     step_size=step_size, minlen=minlen,
                     maxlen=maxlen, random_seed=rng_seed,
                     trx=trx)


def _tracking(tracker, seeds, dg, stopping_criterion, params_img,
              step_size=0.5, minlen=40, maxlen=200,
              random_seed=None, trx=False):
    """
    Helper function
    """
    if len(seeds.shape) == 1:
        seeds = seeds[None, ...]

    tracker = tqdm(tracker(
        dg,
        stopping_criterion,
        seeds,
        params_img.affine,
        step_size=step_size,
        minlen=minlen,
        maxlen=maxlen,
        return_all=False,
        random_seed=random_seed))

    if trx:
        return LazyTractogram(lambda: tracker,
                              affine_to_rasmm=params_img.affine)
    else:
        return StatefulTractogram(tracker, params_img, Space.RASMM)



# This updates usage/kwargs.rst to the latest API kwargs
# Developers can run this after modifying any arguments the user can see

from AFQ.api.utils import AFQclass_doc

prologue = """
.. _kwargs_docs:

The pyAFQ API optional arguments
--------------------------------
You can run pyAFQ on either a subject or participant level
using pyAFQ's API objects, :class:`AFQ.api.group.GroupAFQ`
and :class:`AFQ.api.participant.ParticipantAFQ`. Either way,
these classes take additional optional arguments. These arguments
give the user control over each step of the tractometry pipeline,
allowing customizaiton of tractography, bundle recognition, registration,
etc. Here are all of these arguments and their descriptions, organized
into 5 sections:

"""


def setup(app):
    with open('./source/reference/kwargs.rst', 'w') as ff:
        ff.write(prologue + AFQclass_doc)



"""
===============================================================
How to add new bundles into pyAFQ (Acoustic Radiations Example)
===============================================================

pyAFQ is designed to be customizable and extensible. This example shows how you
can customize it to define a new bundle based on a definition of waypoint and
endpoint ROIs of your design. In this case, we add the acoustic radiations.

We start by importing some of the components that we need for this example and
fixing the random seed for reproducibility

"""

import os.path as op
import plotly
import numpy as np

from AFQ.api.group import GroupAFQ
import AFQ.api.bundle_dict as abd
import AFQ.data.fetch as afd
from AFQ.definitions.image import ImageFile, RoiImage
np.random.seed(1234)


#############################################################################
# Get dMRI data
# ---------------
# We will analyze one subject from the Healthy Brain Network Processed Open
# Diffusion Derivatives dataset (HBN-POD2) [1]_, [2]_. We'll use a fetcher to
# get preprocessed dMRI data for one of the >2,000 subjects in that study. The
# data gets organized into a BIDS-compatible format in the `~/AFQ_data/HBN`
# folder:

study_dir = afd.fetch_hbn_preproc(["NDARAA948VFH"])[1]

#############################################################################
# Define custom `BundleDict` object
# --------------------------------
# The `BundleDict` object holds information about "include" and "exclude" ROIs,
# as well as endpoint ROIS, and whether the bundle crosses the midline. In this
# case, the ROIs are all defined in the MNI template space that is used as the
# default template space in pyAFQ, but, in principle, other template spaces
# could be used.
#
# The ROIs for the case can be downloaded using a custom fetcher which saves
# the ROIs to a folder and creates a dictionary of paths to the ROIs:

ar_rois = afd.read_ar_templates()

bundles = abd.BundleDict({
    "Left Acoustic Radiation": {
        "start": ar_rois["AAL_Thal_L"],
        "end": ar_rois["AAL_TempSup_L"],
        "cross_midline": False,
    },
    "Right Acoustic Radiation": {
        "start": ar_rois["AAL_Thal_R"],
        "end": ar_rois["AAL_TempSup_R"],
        "cross_midline": False
    }
})


#############################################################################
# Define GroupAFQ object
# ----------------------
# HBN POD2 have been processed with qsiprep [3]_. This means that a brain mask
# has already been computer for them. As you can see in other examples, these
# data also have a mapping calculated for them, which can also be incorporated
# into processing. However, in this case, we will let pyAFQ calculate its own
# SyN-based mapping so that the `combine_bundle` method can be used below to
# create a montage visualization.
#
# For tractography, we use CSD-based probabilistic tractography seeding
# extensively (`n_seeds=4` means 81 seeds per voxel!), but only within the ROIs
# and not throughout the white matter. This is controlled by passing
# `"seed_mask": RoiImage()` in the `tracking_params` dict. The custom bundles
# are passed as `bundle_info=bundles`. The call to `my_afq.export_all()`
# initiates the pipeline.

brain_mask_definition = ImageFile(
    suffix="mask",
    filters={'desc': 'brain',
             'space': 'T1w',
             'scope': 'qsiprep'})

my_afq = GroupAFQ(
    bids_path=study_dir,
    preproc_pipeline="qsiprep",
    participant_labels=["NDARAA948VFH"],
    output_dir=op.join(study_dir, "derivatives", "afq_ar"),
    brain_mask_definition=brain_mask_definition,
    tracking_params={"n_seeds": 4,
                     "directions": "prob",
                     "odf_model": "CSD",
                     "seed_mask": RoiImage(use_endpoints=True)},
    bundle_info=bundles)

my_afq.export_all()

#############################################################################
# Interactive bundle visualization
# --------------------------------
# Another way to examine the outputs is to export the individual bundle
# figures, which show the streamlines, as well as the ROIs used to define the
# bundle. This is an html file, which contains an interactive figure that can
# be navigated, zoomed, rotated, etc.

bundle_html = my_afq.export("indiv_bundles_figures")
plotly.io.show(bundle_html["NDARAA948VFH"]["Left Acoustic Radiation"])

#############################################################################
# References
# ----------
# .. [1] Alexander LM, Escalera J, Ai L, et al. An open resource for
#     transdiagnostic research in pediatric mental health and learning
#     disorders. Sci Data. 2017;4:170181.
#
# .. [2] Richie-Halford A, Cieslak M, Ai L, et al. An analysis-ready and quality
#     controlled resource for pediatric brain white-matter research. Scientific
#     Data. 2022;9(1):1-27.
#
# .. [3] Cieslak M, Cook PA, He X, et al. QSIPrep: an integrative platform for
#     preprocessing and reconstructing diffusion MRI data. Nat Methods.
#     2021;18(7):775-778.



import tempfile
import os.path as op
import logging

import numpy as np

import AFQ.viz.utils as vut

from dipy.tracking.streamline import set_number_of_points

try:
    from dipy.viz import window, actor, ui
    from fury.colormap import line_colors
    import IPython.display as display
except (ImportError, ModuleNotFoundError):
    raise ImportError(vut.viz_import_msg_error("fury"))

viz_logger = logging.getLogger("AFQ")


def _inline_interact(scene, inline, interact):
    """
    Helper function to reuse across viz functions
    """
    if interact:
        viz_logger.info("Showing interactive scene...")
        window.show(scene)

    if inline:
        viz_logger.info("Showing inline scene...")
        tdir = tempfile.gettempdir()
        fname = op.join(tdir, "fig.png")
        window.snapshot(scene, fname=fname, size=(1200, 1200))
        display.display_png(display.Image(fname))

    return scene


def visualize_bundles(sft, n_points=None, bundle_dict=None,
                      bundle=None, colors=None,
                      color_by_direction=False,
                      opacity=1.0,
                      line_width=2.0,
                      flip_axes=[False, False, False],
                      figure=None, background=(1, 1, 1), interact=False,
                      inline=False, **kwargs):
    """
    Visualize bundles in 3D using VTK.
    Parameters not described below are extras to conform fury and plotly APIs.

    Parameters
    ----------
    sft : Stateful Tractogram, str
        A Stateful Tractogram containing streamline information
        or a path to a trk file
        In order to visualize individual bundles, the Stateful Tractogram
        must contain a bundle key in it's data_per_streamline which is a list
        of bundle `'uid'`.

    n_points : int or None
        n_points to resample streamlines to before plotting. If None, no
        resampling is done.

    bundle_dict : dict, optional
        Keys are names of bundles and values are dicts that specify them.
        Default: bundles are either not identified, or identified
        only as unique integers in the metadata.

    bundle : str or int, optional
        The name of a bundle to select from among the keys in `bundle_dict`
        or an integer for selection from the sft metadata.

    colors : dict or list
        If this is a dict, keys are bundle names and values are RGB tuples.
        If this is a list, each item is an RGB tuple. Defaults to a list
        with Tableau 20 RGB values if bundle_dict is None, or dict from
        bundles to Tableau 20 RGB values if bundle_dict is not None.

    color_by_direction : bool
        Whether to color by direction instead of by bundle. Default: False

    opacity : float
        Float between 0 and 1 defining the opacity of the bundle.
        Default: 1.0

    background : tuple, optional
        RGB values for the background. Default: (1, 1, 1), which is white
        background.

    figure : fury Scene object, optional
        If provided, the visualization will be added to this Scene. Default:
        Initialize a new Scene.

    interact : bool
        Whether to provide an interactive VTK window for interaction.
        Default: False

    inline : bool
        Whether to embed the visualization inline in a notebook. Only works
        in the notebook context. Default: False.

    Returns
    -------
    Fury Scene object
    """

    if figure is None:
        figure = window.Scene()

    figure.SetBackground(background[0], background[1], background[2])

    for (sls, color, name, dimensions) in vut.tract_generator(
            sft, bundle, bundle_dict, colors, n_points):
        sls = list(sls)
        if name == "all_bundles":
            color = line_colors(sls)
        for sl in sls:
            if flip_axes[0]:
                sl[:, 0] = dimensions[0] - sl[:, 0]
            if flip_axes[1]:
                sl[:, 1] = dimensions[1] - sl[:, 1]
            if flip_axes[2]:
                sl[:, 2] = dimensions[2] - sl[:, 2]

        if color_by_direction:
            sl_actor = actor.line(sls, opacity=opacity)
        else:
            sl_actor = actor.line(sls, color, opacity=opacity)
        figure.add(sl_actor)
        sl_actor.GetProperty().SetRenderLinesAsTubes(1)
        sl_actor.GetProperty().SetLineWidth(line_width)

    return _inline_interact(figure, inline, interact)


def scene_rotate_forward(scene):
    scene.elevation(90)
    scene.set_camera(view_up=(0.0, 0.0, 1.0))
    scene.reset_camera()
    return scene


def create_gif(figure,
               file_name,
               n_frames=60,
               zoom=1,
               z_offset=0.5,
               size=(600, 600),
               rotate_forward=True):
    """
    Convert a Fury Scene object into a gif

    Parameters
    ----------
    figure: Fury Scene object
        Scene to be converted to a gif

    file_name: str
        File to save gif to.

    n_frames: int, optional
        Number of frames in gif.
        Will be evenly distributed throughout the rotation.
        Default: 60

    zoom: int, optional
        How much to magnify the figure in the fig.
        Default: 1

    size: tuple, optional
        Size of the gif.
        Default: (600, 600)

    rotate_forward: bool, optional
        Whether to rotate the figure forward before converting to a gif.
        Generally necessary for fury scenes.
        Default: True
    """
    if rotate_forward:
        figure = scene_rotate_forward(figure)

    tdir = tempfile.gettempdir()
    window.record(figure, az_ang=360.0 / n_frames, n_frames=n_frames,
                  path_numbering=True, out_path=tdir + '/tgif',
                  magnification=zoom,
                  size=size)

    vut.gif_from_pngs(tdir, file_name, n_frames,
                      png_fname="tgif", add_zeros=True)


def visualize_roi(roi, affine_or_mapping=None, static_img=None,
                  roi_affine=None, static_affine=None, reg_template=None,
                  name='ROI', figure=None, color=np.array([1, 0, 0]),
                  flip_axes=None,
                  opacity=1.0, inline=False, interact=False):
    """
    Render a region of interest into a VTK viz as a volume

    Parameters
    ----------
    roi : str or Nifti1Image
        The ROI information

    affine_or_mapping : ndarray, Nifti1Image, or str, optional
       An affine transformation or mapping to apply to the ROIs before
       visualization. Default: no transform.

    static_img: str or Nifti1Image, optional
        Template to resample roi to.
        Default: None

    roi_affine: ndarray, optional
        Default: None

    static_affine: ndarray, optional
        Default: None

    reg_template: str or Nifti1Image, optional
        Template to use for registration.
        Default: None

    name: str, optional
        Name of ROI for the legend.
        Default: 'ROI'

    color : ndarray, optional
        RGB color for ROI.
        Default: np.array([1, 0, 0])

    flip_axes : None
        This parameter is to conform fury and plotly APIs.

    opacity : float, optional
        Opacity of ROI.
        Default: 1.0

    figure : fury Scene object, optional
        If provided, the visualization will be added to this Scene. Default:
        Initialize a new Scene.

    interact : bool
        Whether to provide an interactive VTK window for interaction.
        Default: False

    inline : bool
        Whether to embed the visualization inline in a notebook. Only works
        in the notebook context. Default: False.

    Returns
    -------
    Fury Scene object
    """
    roi = vut.prepare_roi(roi, affine_or_mapping, static_img,
                          roi_affine, static_affine, reg_template)
    for i, flip in enumerate(flip_axes):
        if flip:
            roi = np.flip(roi, axis=i)
    if figure is None:
        figure = window.Scene()

    roi_actor = actor.contour_from_roi(roi, color=color, opacity=opacity)
    figure.add(roi_actor)

    return _inline_interact(figure, inline, interact)


def visualize_volume(volume, x=None, y=None, z=None, figure=None,
                     flip_axes=None,
                     opacity=0.6, inline=True, interact=False):
    """
    Visualize a volume

    Parameters
    ----------
    volume : ndarray or str
        3d volume to visualize.

    figure : fury Scene object, optional
        If provided, the visualization will be added to this Scene. Default:
        Initialize a new Scene.

    flip_axes : None
        This parameter is to conform fury and plotly APIs.

    opacity : float, optional
        Initial opacity of slices.
        Default: 0.6

    interact : bool
        Whether to provide an interactive VTK window for interaction.
        Default: False

    inline : bool
        Whether to embed the visualization inline in a notebook. Only works
        in the notebook context. Default: False.

    Returns
    -------
    Fury Scene object
    """
    volume = vut.load_volume(volume)

    if figure is None:
        figure = window.Scene()

    shape = volume.shape
    image_actor_z = actor.slicer(volume)
    slicer_opacity = opacity
    image_actor_z.opacity(slicer_opacity)

    image_actor_x = image_actor_z.copy()
    if x is None:
        x = int(np.round(shape[0] / 2))
    image_actor_x.display_extent(x,
                                 x,
                                 0,
                                 shape[1] - 1,
                                 0,
                                 shape[2] - 1)

    image_actor_y = image_actor_z.copy()

    if y is None:
        y = int(np.round(shape[1] / 2))
    image_actor_y.display_extent(0,
                                 shape[0] - 1,
                                 y,
                                 y,
                                 0,
                                 shape[2] - 1)

    figure.add(image_actor_z)
    figure.add(image_actor_x)
    figure.add(image_actor_y)

    show_m = window.ShowManager(figure, size=(1200, 900))
    show_m.initialize()

    if interact:
        line_slider_z = ui.LineSlider2D(min_value=0,
                                        max_value=shape[2] - 1,
                                        initial_value=shape[2] / 2,
                                        text_template="{value:.0f}",
                                        length=140)

        line_slider_x = ui.LineSlider2D(min_value=0,
                                        max_value=shape[0] - 1,
                                        initial_value=shape[0] / 2,
                                        text_template="{value:.0f}",
                                        length=140)

        line_slider_y = ui.LineSlider2D(min_value=0,
                                        max_value=shape[1] - 1,
                                        initial_value=shape[1] / 2,
                                        text_template="{value:.0f}",
                                        length=140)

        opacity_slider = ui.LineSlider2D(min_value=0.0,
                                         max_value=1.0,
                                         initial_value=slicer_opacity,
                                         length=140)

        def change_slice_z(slider):
            z = int(np.round(slider.value))
            image_actor_z.display_extent(
                0, shape[0] - 1, 0, shape[1] - 1, z, z)

        def change_slice_x(slider):
            x = int(np.round(slider.value))
            image_actor_x.display_extent(
                x, x, 0, shape[1] - 1, 0, shape[2] - 1)

        def change_slice_y(slider):
            y = int(np.round(slider.value))
            image_actor_y.display_extent(
                0, shape[0] - 1, y, y, 0, shape[2] - 1)

        def change_opacity(slider):
            slicer_opacity = slider.value
            image_actor_z.opacity(slicer_opacity)
            image_actor_x.opacity(slicer_opacity)
            image_actor_y.opacity(slicer_opacity)

        line_slider_z.on_change = change_slice_z
        line_slider_x.on_change = change_slice_x
        line_slider_y.on_change = change_slice_y
        opacity_slider.on_change = change_opacity

        def build_label(text):
            label = ui.TextBlock2D()
            label.message = text
            label.font_size = 18
            label.font_family = 'Arial'
            label.justification = 'left'
            label.bold = False
            label.italic = False
            label.shadow = False
            label.background = (0, 0, 0)
            label.color = (1, 1, 1)

            return label

        line_slider_label_z = build_label(text="Z Slice")
        line_slider_label_x = build_label(text="X Slice")
        line_slider_label_y = build_label(text="Y Slice")
        opacity_slider_label = build_label(text="Opacity")

        panel = ui.Panel2D(size=(300, 200),
                           color=(1, 1, 1),
                           opacity=0.1,
                           align="right")
        panel.center = (1030, 120)

        panel.add_element(line_slider_label_x, (0.1, 0.75))
        panel.add_element(line_slider_x, (0.38, 0.75))
        panel.add_element(line_slider_label_y, (0.1, 0.55))
        panel.add_element(line_slider_y, (0.38, 0.55))
        panel.add_element(line_slider_label_z, (0.1, 0.35))
        panel.add_element(line_slider_z, (0.38, 0.35))
        panel.add_element(opacity_slider_label, (0.1, 0.15))
        panel.add_element(opacity_slider, (0.38, 0.15))

        show_m.scene.add(panel)

        global size
        size = figure.GetSize()

        def win_callback(obj, event):
            global size
            if size != obj.GetSize():
                size_old = size
                size = obj.GetSize()
                size_change = [size[0] - size_old[0], 0]
                panel.re_align(size_change)

    show_m.initialize()

    figure.zoom(1.5)
    figure.reset_clipping_range()

    if interact:
        show_m.add_window_callback(win_callback)
        show_m.render()
        show_m.start()

    return _inline_interact(figure, inline, interact)


def _draw_core(sls, n_points, figure, bundle_name, indiv_profile,
               labelled_points, dimensions, flip_axes):
    fgarray = np.asarray(set_number_of_points(sls, n_points))
    fgarray = np.median(fgarray, axis=0)

    colormap = np.asarray([
        [0.265625, 0.00390625, 0.328125],
        [0.28125, 0.15625, 0.46875],
        [0.2421875, 0.28515625, 0.53515625],
        [0.19140625, 0.40625, 0.5546875],
        [0.1484375, 0.5078125, 0.5546875],
        [0.12109375, 0.6171875, 0.53515625],
        [0.20703125, 0.71484375, 0.47265625],
        [0.4296875, 0.8046875, 0.34375],
        [0.70703125, 0.8671875, 0.16796875],
        [0.98828125, 0.90234375, 0.14453125]])
    xp = np.linspace(
        np.min(indiv_profile),
        np.max(indiv_profile),
        num=len(colormap))
    line_color = np.ones((n_points, 3))
    for i in range(3):
        line_color[:, i] = np.interp(indiv_profile, xp, colormap[:, i])
    line_color_untouched = line_color.copy()
    for i in range(n_points):
        if i < n_points - 1:
            direc = fgarray[i + 1] - fgarray[i]
            direc = direc / np.linalg.norm(direc)
            light_direc = -fgarray[i] / np.linalg.norm(fgarray[i])
            direc_adjust = np.dot(direc, light_direc)
            direc_adjust = (direc_adjust + 3) / 4
        line_color[i, 0:3] = line_color[i, 0:3] * direc_adjust
    text = [None] * n_points
    for label in labelled_points:
        if label == -1:
            text[label] = str(n_points)
        else:
            text[label] = str(label)

    if flip_axes[0]:
        fgarray[:, 0] = dimensions[0] - fgarray[:, 0]
    if flip_axes[1]:
        fgarray[:, 1] = dimensions[1] - fgarray[:, 1]
    if flip_axes[2]:
        fgarray[:, 2] = dimensions[2] - fgarray[:, 2]

    sl_actor = actor.line([fgarray], line_color)
    figure.add(sl_actor)
    sl_actor.GetProperty().SetRenderLinesAsTubes(1)
    sl_actor.GetProperty().SetLineWidth(20)

    return line_color_untouched


def single_bundle_viz(indiv_profile, sft,
                      bundle, scalar_name,
                      bundle_dict=None,
                      flip_axes=[False, False, False],
                      labelled_nodes=[0, -1],
                      figure=None,
                      include_profile=False):
    """
    Visualize a single bundle in 3D with core bundle and associated profile

    Parameters
    ----------
    indiv_profile : ndarray
        A numpy array containing a tract profile for this bundle for a scalar.

    sft : Stateful Tractogram, str
        A Stateful Tractogram containing streamline information.
        If bundle is an int, the Stateful Tractogram
        must contain a bundle key in it's data_per_streamline which is a list
        of bundle `'uid'.
        Otherwise, the entire Stateful Tractogram will be used as the bundle
        for the visualization.

    bundle : str or int
        The name of the bundle to be used as the label for the plot,
        and for selection from the sft metadata.

    scalar_name : str
        The name of the scalar being used.

    bundle_dict : dict, optional
        This parameter is used if bundle is an int.
        Keys are names of bundles and values are dicts that specify them.
        Default: Either the entire sft is treated as a bundle,
        or identified only as unique integers in the metadata.

    flip_axes : ndarray
        Which axes to flip, to orient the image as RAS, which is how we
        visualize.
        For example, if the input image is LAS, use [True, False, False].
        Default: [False, False, False]

    labelled_nodes : list or ndarray
        Which nodes to label. -1 indicates the last node.
        Default: [0, -1]

    figure : Plotly Figure object, optional
        If provided, the visualization will be added to this Figure. Default:
        Initialize a new Figure.

    include_profile : bool, optional
        Not yet implemented in fury. Default: False

    Returns
    -------
    Fury Figure object
    """
    if figure is None:
        figure = window.Scene()
        figure.SetBackground(1, 1, 1)

    n_points = len(indiv_profile)
    sls, _, bundle_name, dimensions = next(vut.tract_generator(
        sft, bundle, bundle_dict, None, n_points))

    _draw_core(
        sls, n_points, figure, bundle_name, indiv_profile,
        labelled_nodes, dimensions, flip_axes)

    return figure



from AFQ.utils.path import drop_extension
import os.path as op
import inspect

__all__ = ["get_fname", "with_name", "get_base_fname"]


def get_base_fname(output_dir, dwi_data_file):
    # setup up path and base file name for outputs
    # remove suffix and desc from dwi data file name
    used_key_list = ["desc", "space", "to", "from"]
    dwi_dfile_no_ext = op.join(
        output_dir,
        drop_extension(op.basename(dwi_data_file)))
    fname = op.dirname(dwi_dfile_no_ext) + "/"
    for key_val_pair in op.basename(dwi_dfile_no_ext).split("_"):
        if "-" in key_val_pair:
            key = key_val_pair.split("-")[0]
            if key not in used_key_list:
                fname = fname + key_val_pair + "_"
    fname = fname[:-1]
    return fname


def get_fname(base_fname, suffix,
              tracking_params=None, segmentation_params=None):
    fname = base_fname
    if tracking_params is not None and 'odf_model' in tracking_params:
        odf_model = tracking_params['odf_model']
        directions = tracking_params['directions']
        fname = fname + (
            f'_coordsys-RASMM_trkmethod-{directions+odf_model}'
        )
    if segmentation_params is not None:
        fname = fname + f"_recogmethod-AFQ"

    return fname + suffix


# Turn list of tasks into dictionary with names for each task
def with_name(task_list):
    return {f"{task.function.__name__}_res": task for task in task_list}


def get_default_args(func):
    return {
        k: v.default
        for k, v in inspect.signature(func).parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def str_to_desc(string):
    return string.replace("-", "").replace("_", "").replace(" ", "")



import numpy as np
import logging
from time import time

import dipy.tracking.streamline as dts
from dipy.utils.parallel import paramap
from dipy.segment.clustering import QuickBundles
from dipy.segment.metricspeed import AveragePointwiseEuclideanMetric
from dipy.segment.featurespeed import ResampleFeature
from dipy.io.streamline import load_tractogram
from dipy.segment.bundles import RecoBundles
from dipy.io.stateful_tractogram import StatefulTractogram, Space

import AFQ.recognition.utils as abu
import AFQ.recognition.cleaning as abc
import AFQ.recognition.curvature as abv
import AFQ.recognition.roi as abr

bundle_criterion_order = [
    "prob_map", "cross_midline", "start", "end",
    "length", "primary_axis", "include", "exclude",
    "recobundles", "qb_thresh"]


logger = logging.getLogger('AFQ')


def prob_map(b_sls, bundle_def, preproc_imap, prob_threshold, **kwargs):
    b_sls.initiate_selection("Prob. Map")
    # using entire fgarray here only because it is the first step
    fiber_probabilities = dts.values_from_volume(
        bundle_def["prob_map"].get_fdata(),
        preproc_imap["fgarray"], np.eye(4))
    fiber_probabilities = np.mean(fiber_probabilities, -1)
    b_sls.select(
        fiber_probabilities > prob_threshold,
        "Prob. Map")


def cross_midline(b_sls, bundle_def, preproc_imap, **kwargs):
    b_sls.initiate_selection("Cross Mid.")
    accepted = preproc_imap["crosses"][b_sls.selected_fiber_idxs]
    if not bundle_def["cross_midline"]:
        accepted = np.invert(accepted)
    b_sls.select(accepted, "Cross Mid.")


def start(b_sls, bundle_def, preproc_imap, **kwargs):
    accept_idx = b_sls.initiate_selection("Startpoint")
    abr.clean_by_endpoints(
        b_sls.get_selected_sls(),
        bundle_def["start"],
        0,
        tol=preproc_imap["dist_to_atlas"],
        flip_sls=b_sls.sls_flipped,
        accepted_idxs=accept_idx)
    if not b_sls.oriented_yet:
        accepted_idx_flipped = abr.clean_by_endpoints(
            b_sls.get_selected_sls(),
            bundle_def["start"],
            -1,
            tol=preproc_imap["dist_to_atlas"])
        b_sls.reorient(accepted_idx_flipped)
        accept_idx = np.logical_xor(
            accepted_idx_flipped, accept_idx)
    b_sls.select(accept_idx, "Startpoint")


def end(b_sls, bundle_def, preproc_imap, **kwargs):
    accept_idx = b_sls.initiate_selection("endpoint")
    abr.clean_by_endpoints(
        b_sls.get_selected_sls(),
        bundle_def["end"],
        -1,
        tol=preproc_imap["dist_to_atlas"],
        flip_sls=b_sls.sls_flipped,
        accepted_idxs=accept_idx)
    if not b_sls.oriented_yet:
        accepted_idx_flipped = abr.clean_by_endpoints(
            b_sls.get_selected_sls(),
            bundle_def["end"],
            0,
            tol=preproc_imap["dist_to_atlas"])
        b_sls.reorient(accepted_idx_flipped)
        accept_idx = np.logical_xor(
            accepted_idx_flipped, accept_idx)
    b_sls.select(accept_idx, "endpoint")


def length(b_sls, bundle_def, preproc_imap, **kwargs):
    accept_idx = b_sls.initiate_selection("length")
    min_len = bundle_def["length"].get(
        "min_len", 0) / preproc_imap["vox_dim"]
    max_len = bundle_def["length"].get(
        "max_len", np.inf) / preproc_imap["vox_dim"]
    for idx, sl in enumerate(b_sls.get_selected_sls()):
        sl_len = np.sum(
            np.linalg.norm(np.diff(sl, axis=0), axis=1))
        if sl_len >= min_len and sl_len <= max_len:
            accept_idx[idx] = 1
    b_sls.select(accept_idx, "length")


def primary_axis(b_sls, bundle_def, **kwargs):
    b_sls.initiate_selection("orientation")
    accept_idx = abc.clean_by_orientation(
        b_sls.get_selected_sls(),
        bundle_def["primary_axis"],
        bundle_def.get(
            "primary_axis_percentage", None))
    b_sls.select(accept_idx, "orientation")


def include(b_sls, bundle_def, preproc_imap, max_includes,
            parallel_segmentation, **kwargs):
    accept_idx = b_sls.initiate_selection("include")
    flip_using_include = len(bundle_def["include"]) > 1\
        and not b_sls.oriented_yet

    if f'inc_addtol' in bundle_def:
        include_roi_tols = []
        for inc_tol in bundle_def["inc_addtol"]:
            include_roi_tols.append((
                inc_tol / preproc_imap["vox_dim"] + preproc_imap["tol"])**2)
    else:
        include_roi_tols = [preproc_imap["tol"]**2] * len(
            bundle_def["include"])

    include_rois = []
    for include_roi in bundle_def["include"]:
        include_rois.append(np.array(
            np.where(include_roi.get_fdata())).T)

    # with parallel segmentation, the first for loop will
    # only collect streamlines and does not need tqdm
    if parallel_segmentation["engine"] != "serial":
        inc_results = paramap(
            abr.check_sl_with_inclusion, b_sls.get_selected_sls(),
            func_args=[
                include_rois, include_roi_tols],
            **parallel_segmentation)

    else:
        inc_results = abr.check_sls_with_inclusion(
            b_sls.get_selected_sls(),
            include_rois,
            include_roi_tols)

    roi_dists = -np.ones(
        (len(b_sls), max_includes),
        dtype=np.int32)
    if flip_using_include:
        to_flip = np.ones_like(accept_idx, dtype=np.bool8)
    for sl_idx, inc_result in enumerate(inc_results):
        sl_accepted, sl_dist = inc_result

        if sl_accepted:
            if len(sl_dist) > 1:
                roi_dists[sl_idx, :len(sl_dist)] = [
                    np.argmin(dist, 0)[0]
                    for dist in sl_dist]
                first_roi_idx = roi_dists[sl_idx, 0]
                last_roi_idx = roi_dists[
                    sl_idx, len(sl_dist) - 1]
                # Only accept SLs that, when cut, are meaningful
                if (len(sl_dist) < 2) or abs(
                        first_roi_idx - last_roi_idx) > 1:
                    # Flip sl if it is close to second ROI
                    # before its close to the first ROI
                    if flip_using_include:
                        to_flip[sl_idx] =\
                            first_roi_idx > last_roi_idx
                        if to_flip[sl_idx]:
                            roi_dists[sl_idx, :len(sl_dist)] =\
                                np.flip(roi_dists[
                                    sl_idx, :len(sl_dist)])
                    accept_idx[sl_idx] = 1
            else:
                accept_idx[sl_idx] = 1
    # see https://github.com/joblib/joblib/issues/945
    if (
        (parallel_segmentation.get(
            "engine", "joblib") != "serial")
        and (parallel_segmentation.get(
            "backend", "loky") == "loky")):
        from joblib.externals.loky import get_reusable_executor
        get_reusable_executor().shutdown(wait=True)
    b_sls.roi_dists = roi_dists
    if flip_using_include:
        b_sls.reorient(to_flip)
    b_sls.select(accept_idx, "include")


def curvature(b_sls, bundle_def, mapping, img, save_intermediates, **kwargs):
    '''
    Filters streamlines by how well they match
    a curve in orientation and shape but not scale
    '''
    accept_idx = b_sls.initiate_selection("curvature")
    if "sft" in bundle_def["curvature"]:
        ref_sl = bundle_def["curvature"]["sft"]
    else:
        ref_sl = load_tractogram(
            bundle_def["curvature"]["path"], "same",
            bbox_valid_check=False)
    moved_ref_sl = abu.move_streamlines(
        ref_sl, "subject", mapping, img,
        save_intermediates=save_intermediates)
    moved_ref_sl.to_vox()
    moved_ref_sl = moved_ref_sl.streamlines[0]
    moved_ref_curve = abv.sl_curve(
        moved_ref_sl,
        len(moved_ref_sl))
    ref_curve_threshold = np.radians(bundle_def["curvature"].get(
        "thresh", 10))
    cut = bundle_def["curvature"].get("cut", True)
    for idx, sl in enumerate(b_sls.get_selected_sls(
            cut=cut, flip=True)):
        if len(sl) > 1:
            this_sl_curve = abv.sl_curve(sl, len(moved_ref_sl))
            dist = abv.sl_curve_dist(this_sl_curve, moved_ref_curve)
            if dist <= ref_curve_threshold:
                accept_idx[idx] = 1
    b_sls.select(accept_idx, "curvature", cut=cut)


def exclude(b_sls, bundle_def, preproc_imap, **kwargs):
    accept_idx = b_sls.initiate_selection("exclude")
    if f'exc_addtol' in bundle_def:
        exclude_roi_tols = []
        for exc_tol in bundle_def["exc_addtol"]:
            exclude_roi_tols.append((
                exc_tol / preproc_imap["vox_dim"] + preproc_imap["tol"])**2)
    else:
        exclude_roi_tols = [
            preproc_imap["tol"]**2] * len(bundle_def["exclude"])
    exclude_rois = []
    for exclude_roi in bundle_def["exclude"]:
        exclude_rois.append(np.array(
            np.where(exclude_roi.get_fdata())).T)
    for sl_idx, sl in enumerate(b_sls.get_selected_sls()):
        if abr.check_sl_with_exclusion(
                sl, exclude_rois, exclude_roi_tols):
            accept_idx[sl_idx] = 1
    b_sls.select(accept_idx, "exclude")


def recobundles(b_sls, mapping, bundle_def, reg_template, img, refine_reco,
                save_intermediates, rng, rb_recognize_params, **kwargs):
    b_sls.initiate_selection("Recobundles")
    moved_sl = abu.move_streamlines(
        StatefulTractogram(b_sls.get_selected_sls(), img, Space.VOX),
        "template", mapping, reg_template,
        save_intermediates=save_intermediates).streamlines
    rb = RecoBundles(moved_sl, verbose=True, rng=rng)
    _, rec_labels = rb.recognize(
        bundle_def['recobundles']['sl'],
        **rb_recognize_params)
    if refine_reco:
        _, rec_labels = rb.refine(
            bundle_def['recobundles']['sl'], moved_sl[rec_labels],
            **rb_recognize_params)
    if not b_sls.oriented_yet:
        standard_sl = next(iter(bundle_def['recobundles']['centroid']))
        oriented_idx = abu.orient_by_streamline(
            moved_sl[rec_labels],
            standard_sl)
        b_sls.reorient(rec_labels[oriented_idx])
    b_sls.select(rec_labels, "Recobundles")


def qb_thresh(b_sls, bundle_def, preproc_imap, clip_edges, **kwargs):
    b_sls.initiate_selection("qb_thresh")
    cut = clip_edges or ("bundlesection" in bundle_def)
    qbx = QuickBundles(
        bundle_def["qb_thresh"] / preproc_imap["vox_dim"],
        AveragePointwiseEuclideanMetric(
            ResampleFeature(nb_points=12)))
    clusters = qbx.cluster(b_sls.get_selected_sls(
        cut=cut, flip=True))
    cleaned_idx = clusters[np.argmax(
        clusters.clusters_sizes())].indices
    b_sls.select(cleaned_idx, "qb_thresh", cut=cut)


def mahalanobis(b_sls, bundle_def, clip_edges, cleaning_params, **kwargs):
    b_sls.initiate_selection("Mahalanobis")
    clean_params = bundle_def.get("mahal", {})
    clean_params = {
        **cleaning_params,
        **clean_params}
    clean_params["return_idx"] = True
    cut = clip_edges or ("bundlesection" in bundle_def)
    _, cleaned_idx = abc.clean_bundle(
        b_sls.get_selected_sls(cut=cut, flip=True),
        **clean_params)
    b_sls.select(cleaned_idx, "Mahalanobis", cut=cut)


def run_bundle_rec_plan(
        bundle_dict, tg, mapping, img, reg_template, preproc_imap,
        bundle_name, bundle_idx, bundle_to_flip, bundle_roi_dists,
        bundle_decisions,
        **segmentation_params):
    # Warp ROIs
    logger.info(f"Preparing ROIs for {bundle_name}")
    start_time = time()
    bundle_def = dict(bundle_dict.get_b_info(bundle_name))
    bundle_def.update(bundle_dict.transform_rois(
        bundle_name,
        mapping,
        img.affine,
        apply_to_recobundles=True))
    logger.info(f"Time to prep ROIs: {time()-start_time}s")

    b_sls = abu.SlsBeingRecognized(
        tg.streamlines, logger,
        segmentation_params["save_intermediates"],
        bundle_name,
        img, len(bundle_def.get("include", [])))

    inputs = {}
    inputs["b_sls"] = b_sls
    inputs["preproc_imap"] = preproc_imap
    inputs["bundle_def"] = bundle_def
    inputs["max_includes"] = bundle_dict.max_includes
    inputs["mapping"] = mapping
    inputs["img"] = img
    inputs["reg_template"] = reg_template
    for key, value in segmentation_params.items():
        inputs[key] = value

    for criterion in bundle_criterion_order:
        if b_sls and criterion in bundle_def:
            inputs[criterion] = globals()[criterion](**inputs)
    if b_sls:
        mahalanobis(**inputs)

    if b_sls and not b_sls.oriented_yet:
        raise ValueError(
            "pyAFQ was unable to consistently orient streamlines "
            f"in bundle {bundle_name} using the provided ROIs. "
            "This can be fixed by including at least 2 "
            "waypoint ROIs, or by using "
            "endpoint ROIs.")

    if b_sls:
        bundle_to_flip[
            b_sls.selected_fiber_idxs,
            bundle_idx] = b_sls.sls_flipped.copy()
        bundle_decisions[
            b_sls.selected_fiber_idxs,
            bundle_idx] = 1
        if hasattr(b_sls, "roi_dists"):
            bundle_roi_dists[
                b_sls.selected_fiber_idxs,
                bundle_idx
            ] = b_sls.roi_dists.copy()



from AFQ.viz.utils import COLOR_DICT
import numpy as np
import scipy.stats as stats
import altair as alt


def altair_color_dict(names_to_include=None):
    """
    Given a list of bundle names, return a dictionary of colors for each
    Formatted for Altair.
    """
    altair_cd = dict(COLOR_DICT.copy())
    for key in list(altair_cd.keys()):
        value = altair_cd[key]
        if (names_to_include is None) or (key in names_to_include):
            altair_cd[key] = (
                f"rgb({int(value[0]*255)},"
                f"{int(value[1]*255)},"
                f"{int(value[2]*255)})")
        else:
            del altair_cd[key]
    return altair_cd


def combined_profiles_df_to_altair_df(
        profiles,
        tissue_properties=['dti_fa', 'dti_md']):
    """
    Given a profiles dataframe that is combined
    from many subjects, return a dataframe formatted for Altair.
    """
    profiles = profiles.copy()
    if 'dki_md' in tissue_properties:
        profiles.dki_md = profiles.dki_md * 1000.
    if 'dti_md' in tissue_properties:
        profiles.dti_md = profiles.dti_md * 1000.

    id_vars = ['tractID', 'nodeID', 'subjectID']
    if 'sessionID' in profiles.columns:
        id_vars.append('sessionID')

    profiles = profiles.melt(
        id_vars=id_vars,
        value_vars=tissue_properties,
        var_name='TP',
        value_name='Value')

    # Function to calculate 95% CI using a normal distribution
    def calculate_95CI(x):
        ci = stats.norm.interval(
            0.95, loc=np.mean(x), scale=np.std(x) / np.sqrt(len(x)))
        return ci

    # Group by 'tractID', 'nodeID', 'TP' and apply the aggregation functions
    profiles = profiles.groupby(['tractID', 'nodeID', 'TP'])['Value'].agg(
        mean='mean',
        CI_lower=lambda x: calculate_95CI(x)[0],
        CI_upper=lambda x: calculate_95CI(x)[1],
        IQR_lower=lambda x: x.quantile(0.25),
        IQR_upper=lambda x: x.quantile(0.75)
    ).reset_index()

    def get_hemi(cc):
        if cc == "L":
            return "Left"
        elif cc == "R":
            return "Right"
        else:
            return "Callosal"

    def get_bname(s):
        if s.startswith("Left "):
            return s[5:]
        elif s.startswith("Right "):
            return s[6:]
        return s

    def formal_tp(tp_name):
        return tp_name.upper().replace("_", " ")

    profiles["Hemi"] = profiles["tractID"].apply(lambda x: get_hemi(x[-1]))
    profiles["Bundle Name"] = profiles["tractID"].apply(get_bname)
    profiles["TP"] = profiles["TP"].apply(formal_tp)

    return profiles


def altair_df_to_chart(profiles, position_domain=(20, 80),
                       column_count=1, font_size=20,
                       line_size=10, row_label_angle=90,
                       bundle_list=None,
                       legend_line_size=5,
                       alt_x_kwargs={}, alt_y_kwargs={},
                       **kwargs):
    """
    Given a dataframe formatted for Altair, probably from
    combined_profiles_df_to_altair_df, return a chart.

    Example
    -------
        call_results = results[results.Hemi == "Callosal"]
        stand_results = results[results.Hemi != "Callosal"]
        prof_chart = altair_df_to_chart(call_results)
        prof_chart.save("supp_chart_call.png", dpi=300)
        prof_chart = altair_df_to_chart(stand_results,
            column_count=2, color="Hemi")
        prof_chart.save("supp_chart_stand.png", dpi=300)
    """
    this_cd = altair_color_dict(profiles.tractID.unique())

    alt.data_transformers.disable_max_rows()

    profiles = profiles[np.logical_and(
        profiles.nodeID >= position_domain[0],
        profiles.nodeID < position_domain[1])]

    tp_units = {
        "DKI AWF": "",
        "DKI FA": "",
        "DKI MD": " (µm²/ms)",
        "DKI MK": "",
        "DTI FA": "",
        "DTI MD": " (µm²/ms)"}

    if bundle_list is None:
        bundle_list = profiles["Bundle Name"].unique()

    row_charts = []
    for jj, b_name in enumerate(bundle_list):
        row_dataframe = profiles[profiles["Bundle Name"] == b_name]
        charts = []
        for ii, tp in enumerate(sorted(profiles.TP.unique())):
            this_dataframe = row_dataframe[row_dataframe.TP == tp]
            if jj == 0:
                title_name = tp + tp_units[tp]
            else:
                title_name = ""
            if ii == 0:
                y_axis_title = b_name
            else:
                y_axis_title = ""
            if jj == len(profiles["Bundle Name"].unique()) - 1:
                x_axis_title = "Position (%)"
                useXlab = True
            else:
                x_axis_title = ""
                useXlab = False
            y_kwargs = {
                "scale": alt.Scale(zero=False),
                "title": y_axis_title,
                **alt_y_kwargs}
            x_kwargs = {
                "axis": alt.Axis(title=x_axis_title, labels=useXlab),
                **alt_x_kwargs}
            prof_chart = alt.Chart(
                this_dataframe, title=title_name).mark_line(
                    size=line_size).encode(
                y=alt.Y('mean', **y_kwargs),
                x=alt.X('nodeID', **x_kwargs),
                **kwargs)
            prof_chart = prof_chart + alt.Chart(this_dataframe).mark_line(
                size=line_size, opacity=0.5, strokeDash=[1, 1]).encode(
                y=alt.Y('IQR_lower', **y_kwargs),
                x=alt.X('nodeID', **x_kwargs),
                **kwargs)
            prof_chart = prof_chart + alt.Chart(this_dataframe).mark_line(
                size=line_size, opacity=0.5, strokeDash=[1, 1]).encode(
                y=alt.Y('IQR_upper', **y_kwargs),
                x=alt.X('nodeID', **x_kwargs),
                **kwargs)
            prof_chart = prof_chart + alt.Chart(this_dataframe).mark_line(
                size=line_size, opacity=0.5).encode(
                y=alt.Y('CI_lower', **y_kwargs),
                x=alt.X('nodeID', **x_kwargs),
                **kwargs)
            prof_chart = prof_chart + alt.Chart(this_dataframe).mark_line(
                size=line_size, opacity=0.5).encode(
                y=alt.Y('CI_upper', **y_kwargs),
                x=alt.X('nodeID', **x_kwargs),
                **kwargs)
            charts.append(prof_chart)
        row_charts.append(alt.HConcatChart(hconcat=charts))
    return alt.VConcatChart(vconcat=row_charts).configure_axis(
        labelFontSize=font_size,
        titleFontSize=font_size,
        labelLimit=0
    ).configure_legend(
        labelFontSize=font_size,
        titleFontSize=font_size,
        titleLimit=0,
        labelLimit=0,
        columns=column_count,
        symbolStrokeWidth=legend_line_size * 10,
        symbolSize=legend_line_size * 100,
        orient='right'
    ).configure_title(
        fontSize=font_size
    )
